#ifndef ALIGNER_P_HH
#define ALIGNER_P_HH

#include "sequence.hh"
#include "basepairs.hh"
#include "alignment.hh"

#include "params.hh"
#include "scoring.hh"

#include "matrices.hh"

/*
  TODO: score matrices should be as small as possible
  and have offsets.
  The M-Matrix can be smaller if the maximal arc len is
  limited. The D-matrix may be smaller if the number
  of simultaneously needed arc-pairs is limited, due
  to limit on the local sub-sequence lengths.
  
  A special cyclically rotatable matrix is needed for
  alignment on the top-level.
*/

/**
   Contains data for restricting AlignerP to sub-sequences
*/
class AlignerPRestriction {
  typedef size_t size_type;
private:    
    size_type startA;
    size_type startB;
    size_type endA;
    size_type endB;
public:
    AlignerPRestriction(size_type startA_, size_type startB_, size_type endA_, size_type endB_)
	: startA(startA_), startB(startB_), endA(endA_), endB(endB_)
    {}

    size_t get_startA() const {//std::cout<<"StartA "<<startA<<std::endl;
    return startA;}
    size_t get_endA() const {//std::cout<<"endA "<<endA<<std::endl; 
    	return endA;}
    
    size_t get_startB() const {//std::cout<<"StartB "<<startB<<std::endl;
    return startB;}
    size_t get_endB() const {//std::cout<<"endB "<<endB<<std::endl; 
    	return endB;}

    void set_startA(size_t p) {startA=p;}
    void set_endA(size_t p)  {endA=p;}
    void set_startB(size_t p)  {startB=p;}
    void set_endB(size_t p)  {endB=p;}
};

inline
std::ostream & operator<<(std::ostream &out, AlignerPRestriction r) {
    return
	out << r.get_startA() << " "
	    << r.get_startB() << " "
	    << r.get_endA() << " "
	    << r.get_endB();
}

//! an extended pf_score_t that can store and calculate with
//! infinite values (i.p. we use -infty for invalid matrix entries)

typedef Matrix<pf_score_t> PFScoreMatrix;
typedef Matrix<double> ProbMatrix;

/* ============================================================ 
   class AlignerP

   Compute alignment similarity score and optimal alignment
   
   initialize with two sequences and their basepairs (including scores)
   ============================================================ 
*/

#define FOR_IJ_IN_ARCMATCH \
  for (size_type i=al+1; i<ar; i++) \
    for (size_type j=bl+1; j<br; j++)

//! class for doing the alignment of two sequences and their
//! associated sets of weighted basepairs
//!
//! an object always knows about the two sequences and
//! the two weighted base pair sets
//!
//! common usage: construct, align, trace, get_alignment
class AlignerP {
public:
    typedef size_t size_type;
protected:
    const Scoring *scoring; //!< the scores
    
    const AlignerParams *params; //!< the parameter for the alignment
	
    Sequence &seqA;
    BasePairs &bpsA;
    Sequence &seqB;
    BasePairs &bpsB;

    // The following is added for the idea of
    // getting the k-best local alignments
    //
    // AlignerP always works on sub-sequences/structures of
    // seqA and seqB. These are specified by
    AlignerPRestriction r;
    // in the standard case, i.e. align the whole sequences,
    // the values are set to startA=1 and endA=seqA.length()
    // and analogously for B
	
    // the constructor initializes to the above standard values
    // methods provide possibility to restriction
    

    pf_score_t partFunc; //!< the total partition function (only defined after call of align_inside())

    PFScoreMatrix D;
  
    PFScoreMatrix M;

    PFScoreMatrix MPrime;
    PFScoreMatrix DPrime;
    
    PFScoreMatrix Mrev;
    ProbMatrix am_prob;
    ProbMatrix bm_prob;
    size_type max_i,max_j; // computed by align_top_level
    
    bool D_created; // flag, is D already created?
    bool DPrime_created; // flag, is DPrime already created?

    Alignment alignment;

    bool am_heuristic;


    //! initialize first column and row, for inside recursion
    void init_M_inside(size_type al, size_type ar, size_type bl, size_type br);
    
    void reversed_M_init(size_type al, size_type ar, size_type bl, size_type br); 

    //! initialize first column and row, for outside recursion
    void init_MPrime(size_type al, size_type ar, size_type bl, size_type br);
    //! initialize first column and row, for reverted outside recursion
    void init_Mrev(size_type al, size_type ar, size_type bl, size_type br) ;
    // recursion cases that handle everything but exclusions
    // (in the LSSA-paper this function was called NoEx
    
    //! standard cases for alignment returning partition function scores (without exlusion handling)
    //!uses inside recursion
    pf_score_t align_noex_inside(size_type al, size_type bl, size_type i, size_type j);
    //! standard cases for alignment returning partition function scores (without exlusion handling)
    //!uses outside recursion
    pf_score_t align_noex_outside(size_type al, size_type bl, size_type i, size_type j);
    pf_score_t align_noex_reversedM(size_type al, size_type bl, size_type i, size_type j);
    
    //! align, where al,ar,bl,br is an arc-match
    //! in structure local alignment, this allows to introduce exclusions-uses inside recursion
    void align_in_arcmatch(size_type al,size_type ar,size_type bl,size_type br);
    //! align, where al,ar,bl,br is an arc-match
     //! in structure local alignment, this allows to introduce exclusions - uses outside recursion
    void align_out_arcmatch(size_type al,size_type ar,size_type bl,size_type br);
    //! Align, outside in reversed way
    void align_rev(size_type al, size_type bl,size_type i, size_type j);
    
    //! create the entries in the D matrix
    //!   This function is called by align() (unless D_created)
    //! uses inside recursion
    void align_D_inside();
    //! create the entries in the DPrime matrix
    //!   This function is called by align() (unless DPrime_created)
    //! uses inside recursion
    void align_DPrime_outside();
    
    //!  fill in D the entries with left ends al,bl, 
    //! where adjlA, adjlB are adjanceny lists of the arcs
    //! uses inside recursion
    void 
    fill_D_entries_inside(size_type al, size_type bl,
		   const BasePairs::LeftAdjList &adjlA,
		   const BasePairs::LeftAdjList &adjlB,
		   size_type max_ar, size_type max_br);
    //!  fill in D the entries with right ends ar,br, 
    //! where adjrA, adjrB are adjanceny lists of the arcs
    //! uses outside recursion
    void 
    fill_DPrime_entries_outside(size_type al, size_type bl,
				const BasePairs::LeftAdjList &adjlA,
				const BasePairs::LeftAdjList &adjlB,
				size_type min_ar, size_type min_br
				);
    
   
public:  
    //! copy constructor
    // AlignerP(const AlignerP &a);
    
    //! construct with sequences and corresponding sets of basepairs
    AlignerP(Sequence &seqA, BasePairs &bpsA,
	     Sequence &seqB, BasePairs &bpsB,
	     const AlignerParams *ap,
	     const Scoring *s
	    );
        
    //! assignment operator
    // const AlignerP & operator = (const AlignerP &a);
    
    //! compute the partition function by the inside algorithm
    //! and fill the D matrix
    pf_score_t align_inside();
       
    //! perform the outside algorithm
    //! and fill the Dprime matrix,
    //! assumes that D matrix is computed already
    void align_outside();
    
    //! computes the probabilitites of all base matches and stores them internally (in a matrix)
    void compute_basematch_probabilities( bool basematch_probs_include_arcmatch );

    //! computes the probabilitites of all arc matches and stores them internally (in a sparse matrix)
    void compute_arcmatch_probabilities();
  
    void write_arcmatch_probabilities(std::ostream &out);
    void write_basematch_probabilities(std::ostream &out); 
    
    
  //  const Matrix *get_basematch_probabilities(){}
    //const SparseMatrix *get_arcmatch_probabilities(){}
        
};

#endif
