/**********************************************************************
 *
 * LocARNA: LOCal Alignment of RNA
 *
 * Copyright (C) Sebastian Will <will(@)informatik.uni-freiburg.de> 
 *               2005, 2006, 2007, 2008
 *
 **********************************************************************/


#include <iostream>
#include <fstream>
// #include <sstream>
#include <vector>

#include <math.h>

#include "sequence.hh"
#include "basepairs.hh"
#include "alignment.hh"
#include "aligner.hh"

#include "rna_data.hh"

#include "match_probs.hh"
#include "ribosum.hh"

#include "ribosum85_60.icc"

#include "anchor_constraints.hh"

using namespace std;

const std::string 
VERSION_STRING = (std::string)PACKAGE_STRING; 

// ------------------------------------------------------------
// Parameter

double min_prob; // only pairs with a probability of at least min_prob are taken into account

int match_score;
int mismatch_score;
int indel_score;
int indel_opening_score;
int temperature;
int struct_weight;

int tau_factor; // contribution of sequence similarity in an arc match (in percent)

bool no_lonely_pairs; // no lonely pairs option

bool struct_local; // allow exclusions for maximizing alignment of connected substructures 
bool sequ_local; // maximize alignment of subsequences

const bool DO_TRACE=true;


int max_diff; // maximal difference for positions of alignment edges
// (only used for ends of arcs)
int max_diff_am; //maximal difference between two arc ends, -1 is off

// only consider arc matchs where
//   1. for global (bl-al)>max_diff || (br-ar)<=max_diff    (if max_diff>=0)
//   2. for local (ar-al)-(br-bl)<=max_diff_am              (if max_diff_am>=0)
// except when there is no additional computation of M matrices necessary,
// this occurs if arcs are left-incident with larger arcs where 1 and 2 hold


int exclusion_score; // Score contribution per exclusion
// set to zero for unrestricted structure locality

// double prob_exp_f(int seqlen) {return 1.0/(2*seqlen);} // expected probability of a base pair (null-model)

double exp_prob;
bool opt_exp_prob;

int output_width;

// ------------------------------------------------------------
// File arguments
std::string bpsfile1;
std::string bpsfile2;

std::string clustal_out;
bool opt_clustal_out;

std::string pp_out;
bool opt_pp_out;

// ------------------------------------------------------------
//
// Options
//
#include "options.hh"


bool opt_help;
bool opt_version;
bool opt_verbose;
bool opt_local_output;
bool opt_pos_output;

bool opt_stacking;

std::string ribosum_file;
bool use_ribosum;

bool opt_probcons_file;
std::string probcons_file;

bool opt_mea_alignment;

bool opt_write_matchprobs;
bool opt_read_matchprobs;
std::string matchprobs_file;

bool opt_write_arcmatch_scores;
bool opt_read_arcmatch_scores;
std::string arcmatch_scores_file;

int match_prob_method;

double min_am_prob; // only matched arc-pair with a probability of at least min_am_prob are taken into account
double min_bm_prob; // only matched base-pair with a probability of at least min_bm_prob are taken into account
int kbest_k;

std::string seq_constraints_A;
std::string seq_constraints_B;

bool opt_ignore_constraints;

int pf_struct_weight;

int mea_alpha;
int mea_beta;
int mea_gamma;


option_def my_options[] = {
    {"help",'h',&opt_help,O_NO_ARG,0,O_NODEFAULT,"","This help"},
    {"version",'V',&opt_version,O_NO_ARG,0,O_NODEFAULT,"","Version info"},
    {"verbose",'v',&opt_verbose,O_NO_ARG,0,O_NODEFAULT,"","Verbose"},

    {"",0,0,O_SECTION,0,O_NODEFAULT,"","Scoring parameters"},

    {"match",'m',0,O_ARG_INT,&match_score,"50","score","Match score"},
    {"mismatch",'M',0,O_ARG_INT,&mismatch_score,"0","score","Mismatch score"},
    {"ribosum-file",0,0,O_ARG_STRING,&ribosum_file,"RIBOSUM85_60","f","Ribosum file"},
    {"use-ribosum",0,0,O_ARG_BOOL,&use_ribosum,"true","bool","Use ribosum scores"},
    {"indel",'i',0,O_ARG_INT,&indel_score,"-350","score","Indel score"},
    {"indel-opening",0,0,O_ARG_INT,&indel_opening_score,"-500","score","Indel opening score"},
    {"struct-weight",'s',0,O_ARG_INT,&struct_weight,"180","score","Maximal weight of 1/2 arc match"},
    {"exp-prob",'e',&opt_exp_prob,O_ARG_DOUBLE,&exp_prob,O_NODEFAULT,"prob","Expected probability"},
    {"tau",'t',0,O_ARG_INT,&tau_factor,"0","factor","Tau factor in percent"},
    {"exclusion",'E',0,O_ARG_INT,&exclusion_score,"0","score","Exclusion weight"},
    {"stacking",0,&opt_stacking,O_NO_ARG,0,O_NODEFAULT,"","Use stacking terms (needs stack-probs by RNAfold -p2)"},   

    {"",0,0,O_SECTION,0,O_NODEFAULT,"","Type of locality"},

    {"struct-local",0,0,O_ARG_BOOL,&struct_local,"false","bool","Structure local"},
    {"sequ-local",0,0,O_ARG_BOOL,&sequ_local,"false","bool","Sequence local"},

    {"",0,0,O_SECTION,0,O_NODEFAULT,"","Controlling output"},

    {"width",'w',0,O_ARG_INT,&output_width,"120","columns","Output width"},
    {"clustal",0,&opt_clustal_out,O_ARG_STRING,&clustal_out,O_NODEFAULT,"file","Clustal output"},
    {"pp",0,&opt_pp_out,O_ARG_STRING,&pp_out,O_NODEFAULT,"file","PP output"},
    {"local-output",'L',&opt_local_output,O_NO_ARG,0,O_NODEFAULT,"","Output only local sub-alignment"},
    {"pos-output",'P',&opt_pos_output,O_NO_ARG,0,O_NODEFAULT,"","Output only local sub-alignment positions"},

    {"",0,0,O_SECTION,0,O_NODEFAULT,"","Heuristics for speed accuracy trade off"},

    {"min-prob",'p',0,O_ARG_DOUBLE,&min_prob,"0.0005","prob","Minimal probability"},
    {"max-diff-am",'D',0,O_ARG_INT,&max_diff_am,"-1","diff","Maximal difference for sizes of matched arcs"},
    {"min-am-prob",'a',0,O_ARG_DOUBLE,&min_am_prob,"0.0005","amprob","Minimal Arc-match probability"},
    {"min-bm-prob",'b',0,O_ARG_DOUBLE,&min_bm_prob,"0.0005","bmprob","Minimal Base-match probability"},
		   
    {"",0,0,O_SECTION,0,O_NODEFAULT,"","Special sauce options"},
    {"kbest",'k',0,O_ARG_INT,&kbest_k,"-1","k","Find k-best alignments"},

    {"",0,0,O_SECTION,0,O_NODEFAULT,"","Options for controlling MEA score (experimental, under construction)"},

    {"mea-alignment",0,&opt_mea_alignment,O_NO_ARG,0,O_NODEFAULT,"","Do MEA alignment"},
    {"probcons-file",0,&opt_probcons_file,O_ARG_STRING,&probcons_file,O_NODEFAULT,"file","Probcons parameter file"},

    {"match-prob-method",0,0,O_ARG_INT,&match_prob_method,"0","int","Method for computation of match probs"},
    {"temperature",0,0,O_ARG_INT,&temperature,"300","int","Temperature for PF-computation"},
    {"pf-struct-weight",0,0,O_ARG_INT,&pf_struct_weight,"200","weight","Structure weight in PF-computation"},

    {"mea-alpha",0,0,O_ARG_INT,&mea_alpha,"100","weight","Weight alpha for MEA"},
    {"mea-beta",0,0,O_ARG_INT,&mea_beta,"100","weight","Weight beta for MEA"},
    {"mea-gamma",0,0,O_ARG_INT,&mea_gamma,"100","weight","Weight gamma for MEA"},

    {"write-match-probs",0,&opt_write_matchprobs,O_ARG_STRING,&matchprobs_file,O_NODEFAULT,"file","Write match probs to file (don't align!)"},
    {"read-match-probs",0,&opt_read_matchprobs,O_ARG_STRING,&matchprobs_file,O_NODEFAULT,"file","Read match probabilities from file"},

    {"write-arcmatch-scores",0,&opt_write_arcmatch_scores,O_ARG_STRING,&arcmatch_scores_file,O_NODEFAULT,"file","Write arcmatch scores (don't align!)"},
    {"read-arcmatch-scores",0,&opt_read_arcmatch_scores,O_ARG_STRING,&arcmatch_scores_file,O_NODEFAULT,"file","Read arcmatch scores"},

    {"",0,0,O_SECTION,0,O_NODEFAULT,"","Constraints"},

    {"noLP",0,&no_lonely_pairs,O_NO_ARG,0,O_NODEFAULT,"","No lonely pairs"},
    {"anchorA",0,0,O_ARG_STRING,&seq_constraints_A,"","string","Anchor constraints sequence A"},
    {"anchorB",0,0,O_ARG_STRING,&seq_constraints_B,"","string","Anchor constraints sequence B"},
    {"ignore-constraints",0,&opt_ignore_constraints,O_NO_ARG,0,O_NODEFAULT,"","Ignore constraints in pp-file"},

    {"",0,0,O_SECTION_HIDE,0,O_NODEFAULT,"","Experimental options (partially incompleted, immature, no guarantees)"},

    {"max-diff-match",'d',0,O_ARG_INT,&max_diff,"-1","diff","Maximal difference for alignment edges"},

    {"",0,0,O_SECTION,0,O_NODEFAULT,"","RNA sequences and pair probabilities"},

    {"",0,0,O_ARG_STRING,&bpsfile1,O_NODEFAULT,"bps-file 1","Basepairs input file 1"},
    {"",0,0,O_ARG_STRING,&bpsfile2,O_NODEFAULT,"bps-file 2","Basepairs input file 2"},
    {"",0,0,0,0,O_NODEFAULT,"",""}
};


// ------------------------------------------------------------


/* ------------------------------------------------------------
   Experimental feature: adapt score to sequence identity
   ------------------------------------------------------------ */

/*
  prob_equals returns "probability" that two alignment columns are equal
  this function needs to be re-implemented

  double seq_identity(Sequence a, Sequence b) {
  const size_t n=a.length();
  const size_t m=b.length();

  vector<vector<double> > M;

  M.resize(n+1);
  for (size_type i=0; i<=n; i++) M[i].resize(m+1);
        
  for (size_type i=0; i<=n; i++) M[i][0]=0;
  for (size_type j=0; j<=m; j++) M[0][j]=0;

  for (size_type i=1; i<=n; i++)
  for (size_type j=1; j<=m; j++)
  M[i][j] = max(M[i-1][j-1] + a[i].prob_equals(b[j]),
  max(M[i-1][j],
  M[i][j-1]));

  return 2.0*M[n][m]/(double)(n+m);
  }
*/


/* ------------------------------------------------------------ */


// ------------------------------------------------------------
// MAIN

int
main(int argc, char **argv) {

    // ------------------------------------------------------------
    // Process options

    bool process_success=process_options(argc,argv,my_options);

    if (opt_help) {
	cout << VERSION_STRING<<endl;

	cout << "Copyright Sebastian Will, 2004-2008"<<endl<<endl;

	cout << "A tool for pairwise Local (and global) Alignment of RNA."<<endl<<endl;

	print_help(argv[0],my_options);

	cout << "Report bugs to <will (at) informatik.uni-freiburg.de>."<<endl<<endl;
	exit(0);
    }

    if (opt_version || opt_verbose) {
	cout << VERSION_STRING<<endl;
	if (opt_version) exit(0); else cout <<endl;
    }

    if (!process_success) {
	puts(O_error_msg);
	printf("USAGE: ");
	print_usage(argv[0],my_options);
	printf("\n");
	exit(-1);
    }

    if (opt_verbose)
	print_options(my_options);


    // ------------------------------------------------------------
    // Get input data and generate data objects
    //

    BasePairsParams bp_params(min_prob);

    RnaData rnadataA(bpsfile1,opt_stacking);
    RnaData rnadataB(bpsfile2,opt_stacking);

    Sequence seqA=rnadataA.get_sequence();
    Sequence seqB=rnadataB.get_sequence();

    BasePairs bpsA(&rnadataA,&bp_params);
    BasePairs bpsB(&rnadataB,&bp_params);

    // ----------------------------------------
    // report on input in verbose mode
    if (opt_verbose) {
	cout << "Sequence A: "<<endl;
	seqA.write(cout);
	cout<<" (Length:"<< seqA.length()<<", Basepairs:"<<bpsA.num_bps() << ")" <<endl;

	cout << "Sequence B: "<<endl;
	seqB.write(cout);
	cout<<" (Length:"<< seqB.length()<<", Basepairs:"<<bpsB.num_bps() << ")" <<endl;

	// cout << "Base Identity: "<<(seq_identity(seqA,seqB)*100)<<endl; 
    }

    // ----------------------------------------  
    // Ribosum matrix
    //
    Ribosum *ribosum=NULL;

    if (use_ribosum) {
	if (ribosum_file == "RIBOSUM85_60") {
	    if (opt_verbose) {
		std::cout <<"Use built-in ribosum."<<std::endl;
	    }
	    ribosum = new Ribosum85_60();
	} else {
	    ribosum = new RibosumFreq(ribosum_file);
	}
    }

    // ------------------------------------------------------------
    // Match probabilities (for MEA-Alignment)
    //
    MatchProbs *match_probs=0L;

    if (opt_read_matchprobs && !opt_mea_alignment) {
	std::cerr << "Warning: opt_read_matchprobs ignored for non-mea alignment.\n"; 
    }

    if (opt_write_matchprobs || opt_mea_alignment) {
	match_probs = new MatchProbs;

	if (!use_ribosum) {
	    std::cerr << "WARNING: Ribosum scoring used for mea_alignment and computing matchprobs."<<std::endl;
	}

	if (opt_read_matchprobs) {
	    match_probs->read_sparse(matchprobs_file,seqA.length(),seqB.length());
	} else {
	    if (match_prob_method==1) {
		if (!opt_probcons_file) {
		    std::cerr << "Probcons parameter file required for pairHMM-style computation"
			      <<" of basematch probabilities."<<std::endl;
		    print_usage(argv[0],my_options);
		    std::cerr << std::endl;
		    exit(-1);
		}
		if (opt_verbose) {
		    std::cout << "Compute match probabilities using pairHMM."<<std::endl; 
		}

		match_probs->pairHMM_probs(seqA,seqB,probcons_file);
	    } else {
		bool sl=sequ_local;
		if (match_prob_method==2) sl=true;
		if (match_prob_method==3) sl=false;

		if (opt_verbose) {
		    std::cout << "Compute match probabilities using PF sequence alignment."<<std::endl; 
		}

		match_probs->pf_probs(rnadataA,rnadataB,
				      ribosum->get_basematch_scores(),
				      ribosum->alphabet(),
				      indel_opening_score/100.0,
				      indel_score/100.0,
				      pf_struct_weight/100.0,
				      temperature/100.0,
				      sl);
	    }
	}

	if (opt_write_matchprobs) {
	    if (opt_verbose) {
		std::cout << "Write match probabilities to file "<<matchprobs_file<<"."<<std::endl; 
	    }

	    match_probs->write_sparse(matchprobs_file,1.0/Scoring::score_resolution);
	    if (!opt_write_arcmatch_scores) exit(0); // else we exit there!
	}
    }

    // ----------------------------------------
    // construct scoring
    if (opt_verbose) {
	std::cout << "Construct scoring."<<std::endl;
    }

    ScoringParams scoring_params(match_score,
				 mismatch_score,

				 // In true mea alignment gap are only 
				 // scored for computing base match probs.
				 // Consequently, we set the indel and indel opening cost to 0
				 // for the case of mea alignment!
				 opt_mea_alignment?0:indel_score,
				 opt_mea_alignment?0:indel_opening_score,
				 ribosum,
				 struct_weight,
				 tau_factor,
				 exclusion_score,
				 opt_exp_prob?exp_prob:-1,
				 temperature,
				 opt_stacking,
				 opt_mea_alignment,
				 match_probs,
				 mea_alpha,
				 mea_beta,
				 mea_gamma
				 );
	
    Scoring scoring(&bpsA,&bpsB,&scoring_params);
	
    /*
    // ------------------------------------------------------------
    // for mea alignment with probabilistic consistency transformation, handle
    // reading and writing of arcmatch_scores
    //
    if (opt_read_arcmatch_scores) {
    if (opt_verbose) {
    std::cout << "Read arcmatch scores from file " << arcmatch_scores_file << "." <<std::endl;
    }
    scoring.read_arcmatch_scores(arcmatch_scores_file);
    } else if (opt_write_arcmatch_scores) {
    if (opt_verbose) {
    std::cout << "Write arcmatch scores to file "<< arcmatch_scores_file<<" and exit."<<std::endl;
    }
    scoring.write_arcmatch_scores(arcmatch_scores_file);
    exit(0);
    }
    */

    if (opt_verbose) {
	std::cout << "Construct scoring done."<<std::endl;
    }

    // ------------------------------------------------------------
    // Handle constraints (optionally)

    std::string seqCA = seq_constraints_A;
    std::string seqCB = seq_constraints_B;

    if (!opt_ignore_constraints) {
	if ( seqCA=="" ) seqCA = rnadataA.get_seq_constraints();
	if ( seqCB=="" ) seqCB = rnadataB.get_seq_constraints();
    }

    AnchorConstraints seq_constraints(seqA.length(),seqCA,
				      seqB.length(),seqCB);

    if (opt_verbose) {
	if (! seq_constraints.empty()) {
	    std::cout << "Found sequence constraints."<<std::endl;
	}
    }


    // ------------------------------------------------------------
    // Computation of the alignment score
    //

    // parameter for the alignment
    AlignerParams aligner_params(
				 no_lonely_pairs,
				 struct_local, sequ_local,
				 max_diff, max_diff_am,
				 min_am_prob,   min_bm_prob,
				 opt_stacking,
				 seq_constraints
				 );

    // initialize aligner object, which does the alignment computation
    Aligner aligner(seqA,bpsA,seqB,bpsB,&aligner_params,&scoring);


    // if option --kbest is given, use kbest algo (by interval splitting)
    if (kbest_k>0) {
	aligner.kbest(kbest_k,
		      output_width,
		      opt_verbose,
		      opt_local_output,
		      opt_pos_output
		      );
	exit(0);
    }

    // otherwise compute the best alignment
    infty_score_t score = aligner.align();


    // ----------------------------------------
    // report score
    //
    std::cout << "Score: "<<score<<std::endl;


    // ------------------------------------------------------------
    // Traceback
    //
    if (DO_TRACE) {
	    
	    
	if (opt_verbose) {
	    std::cout << "Traceback."<<std::endl;
	}
	    
	aligner.trace();
	
	// for debugging:
	// aligner.get_alignment().write_debug(std::cout);
	
	aligner.get_alignment().write(std::cout, 
				      output_width,
				      score,
				      opt_local_output,
				      opt_pos_output,
				      opt_verbose // write structure if verbose
				      );
	std::cout<<endl;

	// ----------------------------------------
	// optionally write output formats
	//
	if (opt_clustal_out) {
	    ofstream out(clustal_out.c_str());
	    if (out.good()) {
		aligner.get_alignment().write_clustal(out, output_width,
						      score,
						      opt_local_output,
						      opt_pos_output
						      );
	    } else {
		cerr << "Cannot write to "<<clustal_out<<endl<<"! Exit.";
		exit(-1);
	    }
	}
	if (opt_pp_out) {
	    ofstream out(pp_out.c_str());
	    if (out.good()) {
		aligner.get_alignment().
		    write_pp(out,bpsA,bpsB,scoring,seq_constraints,output_width);
	    } else {
		cerr << "Cannot write to "<<pp_out<<endl<<"! Exit.";
		exit(-1);
	    }
	}
    }

    // ----------------------------------------
    // DONE
    exit(0);
}
